/*
 * Copyright (C) 2005-2009 Makoto Sato. All Rights Reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package jp.objectfanatics.assertion.mojo;

import static jp.objectfanatics.commons.javassist.JavassistUtils.getLineNumberEvenIfAvailableLineNumberInfoNotAvailable;
import static jp.objectfanatics.commons.javassist.JavassistUtils.getSourceFileName;

import java.io.File;
import java.io.IOException;
import java.util.List;

import javassist.CtBehavior;
import jp.objectfanatics.assertion.constraints.NotNull;
import jp.objectfanatics.assertion.weaver.ClassFolderWeaver;
import jp.objectfanatics.assertion.weaver.ClassFolderWeaver.WeavingStatusListener;
import jp.objectfanatics.assertion.weaver.Weaver.IllegalParameterOverridingException;
import jp.objectfanatics.assertion.weaver.Weaver.IllegalReturnValueOverridingException;
import jp.objectfanatics.assertion.weaver.Weaver.NoSuitableAnnotatedParameterTypeException;
import jp.objectfanatics.assertion.weaver.Weaver.NoSuitableAnnotatedReturnTypeException;

import org.apache.maven.plugin.AbstractMojo;
import org.apache.maven.plugin.MojoExecutionException;

/**
 * reference
 * <ul>
 *   <li>http://maven.apache.org/plugin-developers/index.html
 *   <li>http://maven.apache.org/guides/introduction/introduction-to-the-lifecycle.html
 * </ul>
 * 
 * @author Makoto Sato
 */
abstract public class AbstractWeavingMojo extends AbstractMojo {
	
	/**
	 * Directory containing the build files.
	 * 
	 * @parameter expression="${project.build.directory}"
	 */
	protected File buildDirectory;
	
	/**
	 * Root path to be woven.
	 */
	protected String path;
	
	/**
	 * Source root path. It may be null.
	 */
	protected String sourceRoot;
	
	@Override
	public void execute() throws MojoExecutionException {
		try {
			String[] classpaths = toArray(getClasspaths());
			getLog().info("start weaving " + path);
			new ClassFolderWeaver(createClassFolderWeaverListener(), classpaths).weave(path);
			getLog().info("end   weaving " + path);
		}
		catch (IOException e) {
			throw new MojoExecutionException(e.getMessage(), e);
		}
	}
	
	private WeavingStatusListener createClassFolderWeaverListener() {
		return new WeavingStatusListener() {
			
			@Override
			public void notified(File classFile, IllegalParameterOverridingException e) {
				int targetMethodLineNumber = getLineNumberEvenIfAvailableLineNumberInfoNotAvailable(e.overridingMethod);
				boolean isLineNumberAvailable = targetMethodLineNumber != -1;
				getLog().info("  - An expected parameter constraint annotation is missing.");
				getLog().info("    parameter index            : " + e.parameterIndex);
				getLog().info("    annotation                 : " + e.annotation.getTypeName());
				getLog().info("    overriding method");
				getLog().info("      source file              : " + createSourceFileName(e.overridingMethod) + (isLineNumberAvailable ? ":" + targetMethodLineNumber : ""));
				if (!isLineNumberAvailable) getLog().info("      line number              : Not Available");
				getLog().info("      class file               : " + classFile.getAbsolutePath());
				getLog().info("      method                   : " + e.overridingMethod.getLongName());
				getLog().info("      parameter name(or index) : " + e.overridingParameterName);
				getLog().info("    overriden method");
				getLog().info("      method                   : " + e.overriddenMethod.getLongName());
				getLog().info("      parameter name(or index) : " + e.overriddenParameterName);
			}
			
			@Override
			public void notified(File classFile, IllegalReturnValueOverridingException e) {
				int targetMethodLineNumber = getLineNumberEvenIfAvailableLineNumberInfoNotAvailable(e.overridingMethod);
				boolean isLineNumberAvailable = targetMethodLineNumber != -1;
				getLog().info("  - An expected return value constraint annotation is missing.");
				getLog().info("    annotation    : " + e.annotation.getTypeName());
				getLog().info("    overriding method");
				getLog().info("      source file : " + createSourceFileName(e.overridingMethod) + (isLineNumberAvailable ? ":" + targetMethodLineNumber : ""));
				if (!isLineNumberAvailable) getLog().info("      line number : Not Available");
				getLog().info("      class file  : " + classFile.getAbsolutePath());
				getLog().info("      method      : " + e.overridingMethod.getLongName());
				getLog().info("    overriden method");
				getLog().info("      method      : " + e.overriddenMethod.getLongName());
			}
			
			@Override
			public void notified(File classFile, NoSuitableAnnotatedParameterTypeException e) {
				int targetMethodLineNumber = getLineNumberEvenIfAvailableLineNumberInfoNotAvailable(e.target);
				boolean isLineNumberAvailable = targetMethodLineNumber != -1;
				getLog().info("  - Illegal use of @" + e.annotation.getSimpleName() + ".");
				getLog().info("    source file              : " + createSourceFileName(e.target) + (isLineNumberAvailable ? ":" + targetMethodLineNumber : ""));
				if (!isLineNumberAvailable) getLog().info("    line number              : Not Available");
				getLog().info("    class file               : " + classFile.getAbsolutePath());
				getLog().info("    annotation               : " + e.annotation.getName());
				getLog().info("    method or constructor    : " + e.target.getLongName());
				getLog().info("    parameter type           : " + e.parameterType.getName());
				getLog().info("    parameter name(or index) : " + e.parameterName);
			}
			
			@Override
			public void notified(File classFile, NoSuitableAnnotatedReturnTypeException e) {
				int targetMethodLineNumber = getLineNumberEvenIfAvailableLineNumberInfoNotAvailable(e.target);
				boolean isLineNumberAvailable = targetMethodLineNumber != -1;
				getLog().info("  - Illegal use of @" + e.annotation.getSimpleName() + ".");
				getLog().info("    source file : " + createSourceFileName(e.target) + (isLineNumberAvailable ? ":" + targetMethodLineNumber : ""));
				if (!isLineNumberAvailable) getLog().info("    line number : Not Available");
				getLog().info("    target method   : " + e.target.getLongName());
				getLog().info("    class file  : " + classFile.getAbsolutePath());
				getLog().info("    annotation  : " + e.annotation.getName());
				getLog().info("    return type : " + e.returnType.getName());
			}
			
			@Override
			public void skipped(File skippedFile) {
				getLog().info("  - skip : " + skippedFile.getAbsolutePath());
			}
		};
	}
	
	/**
	 * Returns classpaths
	 * 
	 * @return classpaths
	 */
	abstract protected List<String> getClasspaths();
	
	/**
	 * Converts List<String> to String[].
	 * 
	 * @param list
	 *            list
	 * @return string array
	 */
	@NotNull
	private String[] toArray(@NotNull List<String> list) {
		String[] classpaths = new String[list.size()];
		for (int i = 0; i < list.size(); i++) {
			classpaths[i] = list.get(i);
		}
		return classpaths;
	}
	
	private String createSourceFileName(CtBehavior targetCtBehavior) {
		if (sourceRoot == null) {
			return getSourceFileName(targetCtBehavior);
		}
		else {
			String packageName = targetCtBehavior.getDeclaringClass().getPackageName();
			return sourceRoot + File.separator + (packageName != null ? (packageName.replace(".", File.separator) + File.separator) : "") + getSourceFileName(targetCtBehavior);
		}
	}
}
